import pytest
import os
import pathlib

from unittest.mock import patch, ANY, mock_open

from gatox.models.workflow import Workflow
from gatox.workflow_parser.utility import check_sus

TEST_WF = """
name: 'Test WF'

on:
  pull_request_target:
  workflow_dispatch:

jobs:
  test:
    runs-on: ['self-hosted']
    steps:

    - name: Execution
      run: |
          echo "Hello World and bad stuff!"
"""

TEST_WF2 = """
name: 'Test WF2'

on:
  pull_request_target:

jobs:
  test:
    runs-on: 'ubuntu-latest'
    steps:
    - name: Execution
      uses: actions/checkout@v4
      with:
        ref: ${{ github.event.pull_request.head.ref }}
"""

TEST_WF3 = """
name: Update Snapshots
on:
  issue_comment:
    types: [created]
jobs:
  updatesnapshots:
    if: ${{ github.event.issue.pull_request && github.event.comment.body == '/update-snapshots'}}
    timeout-minutes: 20
    runs-on: macos-latest
    steps:
      - uses: actions/checkout@v3
        with:
          fetch-depth: 0
          token: ${{ secrets.GITHUB_TOKEN }}
      - name: Get SHA and branch name
        id: get-branch-and-sha
        run: |
          sha_and_branch=$(\
            curl \
              -H 'authorization: Bearer ${{ secrets.GITHUB_TOKEN }}' \
              https://api.github.com/repos/${{ github.repository }}/pulls/${{ github.event.issue.number }} \
            | jq -r '.head.sha," ",.head.ref');
          echo "sha=$(echo $sha_and_branch | cut -d " " -f 1)" >> $GITHUB_OUTPUT
          echo "branch=$(echo $sha_and_branch | cut -d " " -f 2)" >> $GITHUB_OUTPUT
      - name: Fetch Branch
        run: git fetch
      - name: Checkout Branch
        run: git checkout ${{ steps.get-branch-and-sha.outputs.branch }}
      - uses: actions/setup-node@v3
        with:
          node-version: '19'
      - name: Install dependencies
        run: yarn
      - name: Install Playwright browsers
        run: npx playwright install --with-deps chromium
      - name: Update snapshots
        env:
          VITE_TENDERLY_ACCESS_KEY: ${{ secrets.VITE_TENDERLY_ACCESS_KEY }}
          VITE_CHAIN_RPC_URL: ${{ secrets.VITE_CHAIN_RPC_URL }}
          CI: true
        run: npx playwright test --update-snapshots --reporter=list
      - uses: stefanzweifel/git-auto-commit-action@v4
        with:
          commit_message: '[CI] Update Snapshots'
"""

TEST_WF4 = """
name: Benchmarks
on:
  issue_comment:
    types: [created]
jobs:
  check-permission:
    if: github.event.issue.pull_request && startsWith(github.event.comment.body, '/bench')
    runs-on: ubuntu-latest
    steps:
    - name: Check permission
      uses: actions/github-script@v6
      with:
        result-encoding: string
        script: |
          const response = await github.rest.repos.getCollaboratorPermissionLevel({
            owner: context.repo.owner,
            repo: context.repo.repo,
            username: context.actor
          });

          const actorPermissionLevel = response.data.permission;
          console.log(actorPermissionLevel);

          // <- lower higher ->
          // ["none", "read", "write", "admin"]
          if (!(actorPermissionLevel == "admin" || actorPermissionLevel == "write")) {
            core.setFailed("Permission denied.");
          }

  benchmarks:
    # run only when PR comments start with '/bench'.
    if: github.event.issue.pull_request && startsWith(github.event.comment.body, '/bench')
    needs: check-permission
    runs-on: [self-hosted, Linux, X64]
    steps:
    - name: Validate and set inputs
      id: bench-input
      uses: actions/github-script@v6
      with:
        result-encoding: string
        script: |
          const command = `${{ github.event.comment.body }}`.split(" ");
          console.log(command);

          // command should be '/bench chain_name pallets'
          if (command.length != 3) {
            core.setFailed("Invalid input. It should be '/bench [chain_name] [pallets]'");
          }

          core.setOutput("chain", command[1]);
          core.setOutput("pallets", command[2]);

    - name: Free disk space
      run: |
        sudo rm -rf /usr/share/dotnet
        sudo rm -rf /usr/local/lib/android
        sudo rm -rf /opt/ghc
        sudo rm -rf "/usr/local/share/boost"
        sudo rm -rf "$AGENT_TOOLSDIRECTORY"
        df -h

    - name: Get branch and sha
      id: get_branch_sha
      uses: actions/github-script@v6
      with:
        github-token: ${{secrets.GITHUB_TOKEN}}
        result-encoding: string
        script: |
          const pull_request = await github.rest.pulls.get({
            owner: context.repo.owner,
            repo: context.repo.repo,
            pull_number: context.issue.number
          })

          core.setOutput("branch", pull_request.data.head.ref)
          core.setOutput("sha", pull_request.data.head.sha)

    - name: Post starting comment
      uses: actions/github-script@v6
      env:
        MESSAGE: |
          Benchmarks job is scheduled at ${{ github.server_url }}/${{ github.repository }}/actions/runs/${{ github.run_id }}.
          Please wait for a while.
          Branch: ${{ steps.get_branch_sha.outputs.branch }}
          SHA: ${{ steps.get_branch_sha.outputs.sha }}
      with:
        github-token: ${{secrets.GITHUB_TOKEN}}
        result-encoding: string
        script: |
          github.rest.issues.createComment({
            issue_number: context.issue.number,
            owner: context.repo.owner,
            repo: context.repo.repo,
            body: process.env.MESSAGE
          })

    - name: Checkout the source code
      uses: actions/checkout@v3
      with:
        ref: ${{ steps.get_branch_sha.outputs.sha }}
        submodules: true

    - name: Install deps
      run: sudo apt -y install protobuf-compiler

    - name: Install & display rust toolchain
      run: rustup show

    - name: Check targets are installed correctly
      run: rustup target list --installed

    - name: Execute benchmarking
      run: |
        mkdir -p ./benchmark-results
        chmod +x ./scripts/run_benchmarks.sh
        ./scripts/run_benchmarks.sh -o ./benchmark-results -c ${{ steps.bench-input.outputs.chain }} -p ${{ steps.bench-input.outputs.pallets }}
"""

TEST_WF5 = """
name: 'Test WF5'

on:
  pull_request_target:

jobs:
"""


TEST_WF6 = """
name: 'Test WF6'

on:
  pull_request_target:

jobs:
  steps:
    - name: Execution
      uses: actions/checkout@v4
"""

TEST_WF7 = """
name: build

on:
  push:
    branches: [ 'master' ]
  pull_request:
    branches: [ 'master' ]

concurrency:
  group: ${{ github.ref }}-build
  cancel-in-progress: true

jobs:
  build:
    strategy:
      matrix:
        profile: [ 'jdk17', 'jdk17-aarch64' ]
        include:
          - jdk_version: '17'
          - profile: 'jdk17'
            runs_on: ubuntu-latest
          - profile: 'jdk17-aarch64'
            runs_on: [ linux, ARM64 ]
      fail-fast: false

    runs-on: ${{ matrix.runs_on }}

    steps:
      - name: Set up Docker Buildx
        uses: docker/setup-buildx-action@v2
        with:
          driver: docker

"""

DEPENDABOT_FALSE_WF = """
version: 2
updates:
  - package-ecosystem: "github-actions"
    directory: "/"
    schedule:
      interval: "daily"
"""


def test_parse_workflow_db():

    workflow = Workflow("unit_test", DEPENDABOT_FALSE_WF, "main.yml")

    assert workflow.isInvalid() == False


# def test_workflow_write():

#     workflow = Workflow("unit_test", TEST_WF, "main.yml")
#     parser = WorkflowParser(workflow)

#     curr_path = pathlib.Path(__file__).parent.resolve()
#     curr_path = pathlib.Path(__file__).parent.resolve()
#     test_repo_path = os.path.join(curr_path, "files/")

#     with patch("builtins.open", mock_open(read_data="")) as mock_file:
#         parser.output(test_repo_path)

#         mock_file().write.assert_called_once_with(parser.raw_yaml)


# def test_check_injection_no_vulnerable_triggers():
#     workflow = Workflow("unit_test", TEST_WF, "main.yml")
#     parser = WorkflowParser(workflow)

#     with patch.object(parser, "get_vulnerable_triggers", return_value=[]):
#         result = parser.check_injection()
#         assert result == {}


# def test_check_injection_no_job_contents():
#     workflow = Workflow("unit_test", TEST_WF5, "main.yml")
#     parser = WorkflowParser(workflow)

#     result = parser.check_injection()
#     assert result == {}


# def test_check_injection_no_step_contents():
#     workflow = Workflow("unit_test", TEST_WF6, "main.yml")
#     parser = WorkflowParser(workflow)

#     result = parser.check_injection()
#     assert result == {}


# def test_check_injection_comment():
#     workflow = Workflow("unit_test", TEST_WF3, "main.yml")
#     parser = WorkflowParser(workflow)

#     result = parser.check_injection()
#     assert "updatesnapshots" in result


# def test_check_injection_no_tokens():
#     workflow = Workflow("unit_test", TEST_WF, "main.yml")
#     parser = WorkflowParser(workflow)

#     result = parser.check_injection()
#     assert result == {}


# def test_check_pwn_request():
#     workflow = Workflow("unit_test", TEST_WF4, "benchmark.yml")
#     parser = WorkflowParser(workflow)

#     result = parser.check_pwn_request()
#     assert result["candidates"]


# def test_check_sh_runnner():
#     workflow = Workflow("unit_test", TEST_WF7, "build.yml")
#     parser = WorkflowParser(workflow)

#     result = parser.self_hosted()
#     assert len(result) > 0



import pytest
from unittest.mock import patch, MagicMock

from gatox.enumerate.reports.actions import ActionsReport
from gatox.enumerate.results.complexity import Complexity
from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.issue_type import IssueType
from gatox.enumerate.results.result_factory import ResultFactory
from gatox.github.api import Api
from gatox.models.workflow import Workflow
from gatox.workflow_graph.visitors.visitor_utils import VisitorUtils
from gatox.workflow_graph.nodes.workflow import WorkflowNode
from gatox.workflow_graph.nodes.job import JobNode
from gatox.workflow_graph.nodes.step import StepNode
from gatox.workflow_graph.nodes.action import ActionNode
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.graph_builder import WorkflowGraphBuilder


@pytest.fixture
def mock_api():
    return MagicMock(spec=Api)


def test_add_results():
    # Test adding paths to results dict
    path = [WorkflowNode("main", "org/repo1", ".github/workflows/test.yml")]
    results = {}

    VisitorUtils._add_results(path, results, IssueType.ACTIONS_INJECTION)

    assert "org/repo1" in results
    assert len(results["org/repo1"]) == 1


def test_check_mutable_ref():
    # Test immutable refs
    assert not VisitorUtils.check_mutable_ref("github.event.pull_request.head.sha")
    assert not VisitorUtils.check_mutable_ref("github.event.workflow_run.head.sha")
    assert not VisitorUtils.check_mutable_ref("github.sha")
    assert not VisitorUtils.check_mutable_ref("sha", {"pull_request_target"})
    assert not VisitorUtils.check_mutable_ref("github.ref")

    # Test mutable refs
    assert VisitorUtils.check_mutable_ref("ref/heads/main")
    assert VisitorUtils.check_mutable_ref("github.ref||something")


def test_process_context_var():
    # Test processing context variables
    assert VisitorUtils.process_context_var("${{ inputs.test }}") == "test"
    assert VisitorUtils.process_context_var("plain_value") == "plain_value"
    assert VisitorUtils.process_context_var("${{ github.sha }}") == "github.sha"


def test_append_path():
    # Test appending paths
    head = [1, 2, 3]
    tail = [3, 4, 5]
    result = VisitorUtils.append_path(head, tail)
    assert result == [1, 2, 3, 4, 5]

    # Test non-matching paths
    head = [1, 2, 3]
    tail = [4, 5, 6]
    result = VisitorUtils.append_path(head, tail)
    assert result == [1, 2, 3]


@patch("gatox.enumerate.enumerate.Api")
def test_initialize_action_node(mock_api):
    graph = WorkflowGraphBuilder().graph
    api = mock_api  # Mock API if needed
    node = ActionNode("someorg/testaction@v4", "main", "action.yml", "testOrg/repo", {})

    try:
        VisitorUtils.initialize_action_node(graph, api, node)
    except:
        pass  # Expected to fail without proper mocking

    # Just verify tag was removed
    assert "uninitialized" not in node.get_tags()


def test_action_render(capsys, mock_api):

    MOCK_WORKFLOW = """
    on:
      pull_request_target:

    jobs:
      test1:
        runs-on: ubuntu-latest
        steps:
          - name: Test Step
            run: echo ${{github.event.pull_request.title}}

    """

    step_data = {
        "name": "test_step",
        "run": "echo ${{github.event.pull_request.title}}",
    }

    test_workflow = Workflow(
        "testOrg/test_repo", MOCK_WORKFLOW, ".github/workflows/test.yml"
    )
    test_flow = [
        WorkflowNode("main", "testOrg/test_repo", ".github/workflows/test.yml"),
        JobNode("Test1", "main", "testOrg/test_repo", ".github/workflows/test.yml"),
        StepNode(
            step_data,
            "main",
            "testOrg/test_repo",
            ".github/workflows/",
            "tests.yml",
            1,
        ),
    ]

    test_flow[0].initialize(test_workflow)

    flow = ResultFactory.create_injection_result(
        test_flow, Confidence.HIGH, Complexity.ZERO_CLICK
    )

    ActionsReport.render_report(flow)
    captured = capsys.readouterr()

    assert "Repository Name: testOrg/test_repo" in captured.out
    assert "Report Type: InjectionResult" in captured.out
    assert "Context Vars: github.event.pull_request.title" in captured.out



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import logging


from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.complexity import Complexity
from gatox.enumerate.results.issue_type import IssueType
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.graph_builder import WorkflowGraphBuilder
from gatox.workflow_graph.visitors.visitor_utils import VisitorUtils
from gatox.github.api import Api
from gatox.workflow_parser.utility import CONTEXT_REGEX
from gatox.caching.cache_manager import CacheManager


logger = logging.getLogger(__name__)


class DispatchTOCTOUVisitor:
    """
    A visitor class designed to identify Time-Of-Check to Time-Of-Use (TOCTOU) vulnerabilities
    in GitHub workflows that are triggered via workflow dispatch and handle pull request numbers
    without accompanying SHA references.

    TOCTOU vulnerabilities can occur when there is a window of opportunity between the time a
    condition is checked and the time an action is taken based on that condition. This class
    specifically targets workflows that may be susceptible to such vulnerabilities by analyzing
    the paths within the workflow graph.
    """

    @staticmethod
    def find_dispatch_misconfigurations(graph: TaggedGraph, api: Api):
        """
        Identifies TOCTOU vulnerabilities in workflows that are triggered by workflow dispatch
        events and handle pull request (PR) numbers without accompanying SHA references.

        This method performs the following operations:
        1. Retrieves all nodes tagged with "workflow_dispatch" from the workflow graph.
        2. For each of these nodes, it performs a Depth-First Search (DFS) to find paths
           that lead to nodes tagged with "checkout".
        3. Processes each identified path to determine if it contains potential TOCTOU
           vulnerabilities.
        4. Aggregates and renders the results in an ASCII format for easy visualization.

        Args:
            graph (TaggedGraph): The workflow graph containing all nodes and their relationships.
            api (Api): An instance of the API wrapper to interact with GitHub APIs.

        Returns:
            None

        Raises:
            Exception: Catches and logs any exceptions that occur during path processing.
        """
        # Retrieve all nodes tagged with "workflow_dispatch"
        nodes = graph.get_nodes_for_tags(
            [
                "workflow_dispatch",
            ]
        )

        all_paths = []
        results = {}

        # Perform DFS from each "workflow_dispatch" node to find paths to "checkout" nodes
        for cn in nodes:
            try:
                paths = graph.dfs_to_tag(cn, "checkout", api)
                if paths:
                    all_paths.append(paths)
            except Exception as e:
                logger.error(f"Error finding paths for dispatch node: {e}")
                logger.warning(f"Node: {cn}")

        # Process each discovered path to identify TOCTOU vulnerabilities
        for path_set in all_paths:
            for path in path_set:
                try:
                    DispatchTOCTOUVisitor.__process_path(path, graph, api, results)
                except Exception as e:
                    logger.warning(f"Error processing path: {e}")
                    logger.warning(f"Path: {path}")

        return results

    @staticmethod
    def __process_path(path, graph: TaggedGraph, api: Api, results: dict):
        """
        Processes a single path within the workflow graph to identify potential TOCTOU
        vulnerabilities.

        The processing involves:
        - Analyzing each node in the path to extract relevant information.
        - Maintaining lookups for input and environment variables.
        - Checking for mutable references that could lead to vulnerabilities.
        - Aggregating the findings into the results dictionary.

        Args:
            path (list): A list of nodes representing a single path in the workflow graph.
            graph (TaggedGraph): The workflow graph containing all nodes and their relationships.
            api (Api): An instance of the API wrapper to interact with GitHub APIs.
            results (dict): A dictionary to store the findings of the analysis.

        Returns:
            None

        Raises:
            Exception: Propagates any exceptions that occur during the processing of the path.
        """
        input_lookup = {}

        # Workflow dispatch jobs inherently have an approval gate,
        # so only TOCTOU issues can be exploited.
        approval_gate = True
        env_lookup = {}
        flexible_lookup = {}

        for index, node in enumerate(path):
            tags = node.get_tags()

            if "JobNode" in tags:
                if node.outputs:
                    for o_key, val in node.outputs.items():
                        if "env." in val and val not in env_lookup:
                            for key in env_lookup.keys():
                                if key in val:
                                    flexible_lookup[o_key] = env_lookup[key]

            elif "WorkflowNode" in tags:
                if index != 0 and "JobNode" in path[index - 1].get_tags():
                    # Caller job node
                    node_params = path[index - 1].params
                    # Set lookup for input params
                    input_lookup.update(node_params)

                if index == 0:
                    repo = CacheManager().get_repository(node.repo_name())
                    if repo.is_fork():
                        break

                    # If the workflow dispatch node does not have any inputs,
                    # skip the rest of the path.
                    if not node.inputs:
                        break

                    pr_num_found = False
                    # Process inputs to determine if any contain a PR number.
                    # This is a heuristic to identify workflows that are taking a PR number
                    # or mutable reference.
                    for key, val in node.inputs.items():
                        if "sha" in key.lower():
                            break
                        elif re.search(
                            r"(?:^|[\b_])(pr|pull|pull_request|pr_number)(?:[\b_]|$)",
                            key,
                            re.IGNORECASE,
                        ):
                            pr_num_found = True
                            break

                    if not pr_num_found:
                        break

                    # Check workflow environment variables for GitHub event references
                    env_vars = node.get_env_vars()
                    for key, val in env_vars.items():
                        if isinstance(val, str):
                            if "github." in val:
                                env_lookup[key] = val

            elif "StepNode" in tags:
                if node.is_checkout:
                    checkout_ref = node.metadata
                    if "inputs." in node.metadata:
                        if "${{" in node.metadata:
                            processed_var = CONTEXT_REGEX.findall(node.metadata)
                            if processed_var:
                                processed_var = processed_var[0]
                                if "inputs." in processed_var:
                                    processed_var = processed_var.replace("inputs.", "")
                            else:
                                processed_var = node.metadata
                        else:
                            processed_var = node.metadata

                        if processed_var in env_lookup:
                            original_val = env_lookup[processed_var]
                            checkout_ref = original_val
                        elif processed_var in input_lookup:
                            checkout_ref = input_lookup[processed_var]

                    if VisitorUtils.check_mutable_ref(checkout_ref):
                        sinks = graph.dfs_to_tag(node, "sink", api)
                        if sinks:
                            VisitorUtils.append_path(path, sinks[0])
                            VisitorUtils._add_results(
                                path,
                                results,
                                IssueType.DISPATCH_TOCTOU,
                                confidence=Confidence.HIGH,
                                complexity=Complexity.TOCTOU,
                            )
                        else:
                            VisitorUtils._add_results(
                                path,
                                results,
                                IssueType.DISPATCH_TOCTOU,
                                confidence=Confidence.LOW,
                                complexity=Complexity.TOCTOU,
                            )

            elif "ActionNode" in tags:
                VisitorUtils.initialize_action_node(graph, api, node)



from unittest.mock import patch

from gatox.attack.runner.webshell import WebShell


@patch("gatox.attack.attack.Api")
def test_shell_workflow_attack(mock_api):
    """ """
    mock_api.return_value.call_get.return_value.status_code = 200
    mock_api.return_value.call_get.return_value.json.return_value = [
        {"tag_name": "v2.317.0"}
    ]

    mock_api.return_value.call_post.return_value.status_code = 201
    mock_api.return_value.call_post.return_value.json.return_value = {
        "token": "LLBF3JGZDX3P5PMEXLND6TS6FCWO6",
        "expires_at": "2020-01-22T12:13:35.123-08:00",
    }

    websheller = WebShell(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
    )

    formatted_gist = websheller.format_ror_gist("c2user/c2repo", "linux", "x64")

    assert formatted_gist is not None
    assert "actions-runner-linux-x64-2.317.0.tar.gz" in formatted_gist



import pytest
from unittest.mock import Mock, patch
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, Future

from gatox.models.repository import Repository

from gatox.enumerate.deep_dive.ingest_non_default import IngestNonDefault

# FILE: gatox/enumerate/deep_dive/test_ingest_non_default.py


@pytest.fixture
def mock_repo():
    repo = Mock(spec=Repository)
    repo.name = "test/repo"
    return repo


@pytest.fixture
def mock_api():
    api = Mock()
    api.pat = "test_pat"
    return api


@patch("gatox.enumerate.deep_dive.ingest_non_default.ThreadPoolExecutor")
def test_ingest_initializes_executor(mock_executor, mock_repo, mock_api):
    IngestNonDefault._executor = None
    IngestNonDefault._api = None
    IngestNonDefault._futures = []

    IngestNonDefault.ingest(mock_repo, mock_api)

    mock_executor.assert_called_once_with(max_workers=8)
    assert IngestNonDefault._api == mock_api


@patch("gatox.enumerate.deep_dive.ingest_non_default.ThreadPoolExecutor")
def test_ingest_submits_task(mock_executor, mock_repo, mock_api):
    mock_executor_instance = mock_executor.return_value
    mock_future = Mock(spec=Future)
    mock_executor_instance.submit.return_value = mock_future

    IngestNonDefault._executor = None
    IngestNonDefault._api = None
    IngestNonDefault._futures = []

    future = IngestNonDefault.ingest(mock_repo, mock_api)

    mock_executor_instance.submit.assert_called_once_with(
        IngestNonDefault._process_repo, mock_repo
    )
    assert future == mock_future
    assert future in IngestNonDefault._futures


@patch("gatox.enumerate.deep_dive.ingest_non_default.Git")
@patch("gatox.enumerate.deep_dive.ingest_non_default.WorkflowGraphBuilder")
def test_process_repo(mock_builder, mock_git, mock_repo, mock_api):
    mock_git_instance = mock_git.return_value
    mock_workflow = Mock()
    mock_git_instance.get_non_default.return_value = [mock_workflow]

    IngestNonDefault._api = mock_api
    IngestNonDefault._process_repo(mock_repo)

    mock_git.assert_called_once_with(mock_api.pat, mock_repo.name)
    mock_git_instance.get_non_default.assert_called_once()
    mock_builder.return_value.build_graph_from_yaml.assert_called_once_with(
        mock_workflow, mock_repo
    )


@patch("gatox.enumerate.deep_dive.ingest_non_default.wait")
def test_pool_empty_no_executor(mock_wait):
    IngestNonDefault._executor = None
    future = IngestNonDefault.pool_empty()

    assert future.done()
    assert future.result() is None
    mock_wait.assert_not_called()


@patch("gatox.enumerate.deep_dive.ingest_non_default.wait")
def test_pool_empty_with_executor(mock_wait):
    mock_future = Mock(spec=Future)
    mock_wait.return_value = ([], [mock_future])

    IngestNonDefault._executor = Mock()
    IngestNonDefault._futures = [mock_future]

    future = IngestNonDefault.pool_empty()

    mock_wait.assert_called_once_with([mock_future], return_when=ALL_COMPLETED)
    assert future.done()
    assert future.result() is None



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
import random
import threading
import logging

from requests.exceptions import RequestException
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

from gatox.caching.cache_manager import CacheManager
from gatox.models.workflow import Workflow
from gatox.models.repository import Repository
from gatox.workflow_graph.graph_builder import WorkflowGraphBuilder
from gatox.cli.output import Output

logger = logging.getLogger(__name__)


class DataIngestor:
    """Utility methods for performing parallel ingestion of data
    from GitHub using threadpools and GraphQL.
    """

    __counter = 0
    __lock = threading.Lock()
    __rl_lock = threading.Lock()

    @classmethod
    def update_count(cls, batch: int):
        """
        Update the class-level counter if the provided batch number is greater than the current counter.

        Args:
            batch (int): The batch number to compare with the current counter.

        Returns:
            None
        """
        with cls.__lock:
            if batch > cls.__counter:
                cls.__counter = batch

    @classmethod
    def check_status(cls):
        """This method returns the value of the class variable __counter."""
        return cls.__counter

    @classmethod
    def perform_parallel_repo_ingest(cls, api, org, repo_count):
        """
        Perform a parallel query of repositories up to the count within a given organization.

        Args:
            api (object): The API client used to make the GET requests.
            org (str): The organization name to query repositories from.
            repo_count (int): The number of repositories to query.

        Returns:
            list: A list of repositories retrieved from the organization.
        """
        repos = []

        def make_query(increment):
            """
            Makes a query to retrieve repositories for a given page.
            Attempts up to 5 times if the request fails.

            Args:
                increment (int): The page number to query.

            Returns:
                list: A list of repositories for the given page, or None if the query fails.
            """
            get_params = {"type": "public", "per_page": 100, "page": increment}

            sleep_timer = 4
            for _ in range(0, 5):
                repos = api.call_get(f"/orgs/{org}/repos", params=get_params)
                if repos.status_code == 200:
                    return repos.json()
                else:
                    time.sleep(sleep_timer)
                    sleep_timer = sleep_timer * 2
            Output.error("Unable to query. Will miss repositories.")

        batches = (repo_count // 100) + 1

        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = []
            for batch in range(1, batches + 1):
                futures.append(executor.submit(make_query, batch))
            for future in as_completed(futures):
                listing = future.result()
                if listing:
                    repos.extend([repo for repo in listing if not repo["archived"]])

        return repos

    @classmethod
    def perform_query(cls, api, work_query, batch):
        """
        Performs a GraphQL query of repositories with up to 3 attempts, increasing the sleep timer from 4, 8, and then finally 16 seconds.

        Args:
            api (object): The API client used to make the POST requests.
            work_query (dict): The GraphQL query to be executed.
            batch (int): The batch number for which the query is being performed.

        Returns:
            list or dict: The nodes or values from the query result if successful, otherwise None.

        Raises:
            Exception: If an error occurs during the query execution.
        """
        try:
            for _ in range(0, 4):
                # We lock if another thread is sleeping due to a rate limit
                # but if no other thread is sleeping, we want to move on
                # to query.
                while cls.__rl_lock.locked():
                    time.sleep(0.1)
                try:
                    result = api.call_post("/graphql", work_query)
                except RequestException:
                    logging.error("Request exception occurred, trying again.")
                    time.sleep(15 + random.randint(0, 3))
                    continue

                # Sometimes we don't get a 200, fall back in this case.
                if result.status_code == 200:
                    json_res = result.json()["data"]
                    DataIngestor.update_count(batch)
                    if "nodes" in json_res:
                        return result.json()["data"]["nodes"]
                    else:
                        return result.json()["data"].values()
                elif result.status_code == 403:
                    with cls.__rl_lock:
                        time.sleep(15 + random.randint(0, 3))
                else:
                    # Add some jitter
                    time.sleep(10 + random.randint(0, 3))

            Output.warn(
                f"GraphQL attempts failed for batch {str(batch)}, will revert to REST for impacted repos."
            )
        except Exception as e:
            Output.warn(
                "Exception while running GraphQL query, will revert to REST "
                "API workflow query for impacted repositories!"
            )
            logger.warning(f"{type(e)}: {str(e)}")

    @staticmethod
    def construct_workflow_cache(yml_results):
        """
        Creates a cache of workflow YAML files retrieved from GraphQL. Since GraphQL and REST do not have parity,
        REST is still used for most enumeration calls. This method saves all YAML files, so during organization-level
        enumeration, if YAML enumeration is performed, the cached file is used instead of making GitHub REST requests.

        Args:
            yml_results (list): List of results from individual GraphQL queries (100 nodes at a time).

        Returns:
            None
        """
        if yml_results is None:
            return

        cache = CacheManager()
        for result in yml_results:
            # If we get any malformed/missing data just skip it and
            # Gato will fall back to the contents API for these few cases.
            if not result:
                continue

            if "nameWithOwner" not in result:
                continue

            owner = result["nameWithOwner"]
            cache.set_empty(owner)
            # Empty means no YAMLs, so just skip.

            default_branch = (
                result["defaultBranchRef"]["name"]
                if result["defaultBranchRef"]
                else "main"
            )

            # If we are using app installation tokens, then
            # the query might return empty for this field, but if
            # we are here then we can read.
            if not result["viewerPermission"]:
                result["viewerPermission"] = "READ"

            repo_data = {
                "full_name": result["nameWithOwner"],
                "html_url": result["url"],
                "visibility": "private" if result["isPrivate"] else "public",
                "default_branch": (
                    result["defaultBranchRef"]["name"]
                    if result["defaultBranchRef"]
                    else "main"
                ),
                "fork": result["isFork"],
                "stargazers_count": result["stargazers"]["totalCount"],
                "pushed_at": result["pushedAt"],
                "permissions": {
                    "pull": result["viewerPermission"] == "READ"
                    or result["viewerPermission"] == "TRIAGE"
                    or result["viewerPermission"] == "WRITE"
                    or result["viewerPermission"] == "MAINTAIN"
                    or result["viewerPermission"] == "ADMIN",
                    "push": result["viewerPermission"] == "WRITE"
                    or result["viewerPermission"] == "MAINTAIN"
                    or result["viewerPermission"] == "ADMIN",
                    "maintain": result["viewerPermission"] == "MAINTAIN"
                    or result["viewerPermission"] == "ADMIN",
                    "admin": result["viewerPermission"] == "ADMIN",
                },
                "archived": result["isArchived"],
                "isFork": result["isFork"],
                "allow_forking": result["forkingAllowed"],
                "environments": [],
            }

            if "environments" in result and result["environments"]:
                # Capture environments not named github-pages
                envs = [
                    env["node"]["name"]
                    for env in result["environments"]["edges"]
                    if env["node"]["name"] != "github-pages"
                ]
                repo_data["environments"] = envs
            repo_wrapper = Repository(repo_data)
            cache.set_repository(repo_wrapper)

            if result["object"]:
                for yml_node in result["object"]["entries"]:
                    yml_name = yml_node["name"]
                    if yml_node["type"] == "blob" and (
                        yml_name.lower().endswith("yml")
                        or yml_name.lower().endswith("yaml")
                    ):
                        if "text" in yml_node["object"]:
                            contents = yml_node["object"]["text"]
                            wf_wrapper = Workflow(
                                owner, contents, yml_name, default_branch=default_branch
                            )

                            if wf_wrapper.isInvalid():
                                continue

                            cache.set_workflow(owner, yml_name, wf_wrapper)
                            WorkflowGraphBuilder().build_graph_from_yaml(
                                wf_wrapper, repo_wrapper
                            )



import unittest
from unittest.mock import patch, mock_open, MagicMock, call
import os
import pathlib
from gatox.cli import cli
from gatox.caching.cache_manager import CacheManager


class TestSaveWorkflowYmls(unittest.TestCase):

    @patch("gatox.cli.cli.CacheManager")
    @patch("pathlib.Path.mkdir")
    @patch("builtins.open", new_callable=mock_open)
    def test_save_workflow_ymls(self, mock_open_func, mock_mkdir, mock_cache_manager):
        """Test that save_workflow_ymls correctly saves workflow files."""

        # Mock CacheManager to return some repos and workflows
        mock_cache_manager_instance = MagicMock()
        mock_cache_manager.return_value = mock_cache_manager_instance

        mock_cache_manager_instance.get_repos.return_value = ["repo1", "repo2"]

        mock_workflow1 = MagicMock(
            workflow_name="wf1.yml", workflow_contents="content1"
        )
        mock_workflow2 = MagicMock(
            workflow_name="wf2.yml", workflow_contents="content2"
        )
        mock_cache_manager_instance.get_workflows.side_effect = [
            [mock_workflow1],
            [mock_workflow2],
        ]

        # Call the function
        output_directory = "/tmp/test_output"
        cli.save_workflow_ymls(output_directory)

        # Assert that mkdir was called correctly
        expected_calls = [
            call(parents=True, exist_ok=True),
            call(parents=True, exist_ok=True),
        ]
        mock_mkdir.assert_has_calls(expected_calls, any_order=True)
        self.assertEqual(mock_mkdir.call_count, 2)

        # Assert that open was called correctly and the contents were written
        mock_open_func.assert_any_call(
            os.path.join(output_directory, "repo1/wf1.yml"), "w"
        )
        mock_open_func.assert_any_call(
            os.path.join(output_directory, "repo2/wf2.yml"), "w"
        )

        handler = mock_open_func()

        handler.write.assert_any_call("content1")
        handler.write.assert_any_call("content2")


if __name__ == "__main__":
    unittest.main()



import requests
import json

from gatox.configuration.configuration_manager import ConfigurationManager


def send_slack_webhook(message):
    # Create the payload
    payload = {"text": json.dumps(message, indent=4)}
    hooks = ConfigurationManager().NOTIFICATIONS["SLACK_WEBHOOKS"]

    for webhook in hooks:
        # Send the request
        response = requests.post(webhook, json=payload)
        # Check the response
        if response.status_code != 200:
            raise ValueError(
                "Request to slack returned an error %s, the response is:\n%s"
                % (response.status_code, response.text)
            )



import os
import pathlib
import pytest
import json
import datetime

from unittest.mock import MagicMock

import gatox.enumerate.repository
from gatox.models.repository import Repository
from gatox.enumerate.repository import RepositoryEnum
from gatox.cli.output import Output

TEST_REPO_DATA = None
TEST_WORKFLOW_YML = None

Output(True)


@pytest.fixture(scope="session", autouse=True)
def load_test_files(request):
    global TEST_REPO_DATA
    global TEST_WORKFLOW_YML
    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/example_repo.json")
    test_wf_path = os.path.join(curr_path, "files/main.yaml")

    with open(test_repo_path, "r") as repo_data:
        TEST_REPO_DATA = json.load(repo_data)

    with open(test_wf_path, "r") as wf_data:
        TEST_WORKFLOW_YML = wf_data.read()


def test_enumerate_repo():
    """Test constructor for enumerator."""
    mock_api = MagicMock()

    gh_enumeration_runner = RepositoryEnum(mock_api, False)

    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.retrieve_run_logs.return_value = [
        {
            "machine_name": "unittest1",
            "runner_name": "much_unit_such_test",
            "runner_type": "organization",
            "non_ephemeral": False,
            "token_permissions": {"Actions": "write"},
            "runner_group": "Default",
            "requested_labels": ["self-hosted", "Linux", "X64"],
        }
    ]

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    test_repo = Repository(repo_data)
    test_repo.add_self_hosted_workflows(["build.yaml"])

    gh_enumeration_runner.enumerate_repository(test_repo)

    assert test_repo.sh_runner_access is True
    assert len(test_repo.accessible_runners) > 0
    assert test_repo.accessible_runners[0].runner_name == "much_unit_such_test"


def test_enumerate_repo_admin():
    """Test constructor for enumerator."""
    mock_api = MagicMock()

    gh_enumeration_runner = RepositoryEnum(mock_api, False)

    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.retrieve_run_logs.return_value = [
        {
            "machine_name": "unittest1",
            "runner_name": "much_unit_such_test",
            "runner_type": "organization",
            "non_ephemeral": False,
            "token_permissions": {"Actions": "write"},
            "runner_group": "Default",
            "requested_labels": ["self-hosted", "Linux", "X64"],
        }
    ]

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    repo_data["permissions"]["admin"] = True
    test_repo = Repository(repo_data)

    gh_enumeration_runner.enumerate_repository(test_repo)

    assert test_repo.is_admin()


def test_enumerate_repo_secrets():
    """Test constructor for enumerator."""
    mock_api = MagicMock()

    gh_enumeration_runner = RepositoryEnum(mock_api, False)

    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.get_secrets.return_value = [
        {
            "name": "GIST_ID",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "private",
        },
        {
            "name": "DEPLOY_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "all",
        },
        {
            "name": "GH_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "selected",
            "selected_repositories_url": "https://api.github.com/orgs/octo-org/actions/secrets/SUPER_SECRET/repositories",
        },
    ]

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    test_repo = Repository(repo_data)

    gh_enumeration_runner.enumerate_repository_secrets(test_repo)

    assert len(test_repo.secrets) > 0



import re
import logging

from datetime import datetime, timedelta

from gatox.configuration.configuration_manager import ConfigurationManager
from gatox.workflow_parser.expression_parser import ExpressionParser
from gatox.workflow_parser.expression_evaluator import ExpressionEvaluator

logger = logging.getLogger(__name__)

pattern = re.compile(
    r"checkout\s+(\$\{\{)?\s*(\S*([a-z$_]+)\S*)\s*(\}\})?", re.IGNORECASE
)

CONTEXT_REGEX = re.compile(r"\${{\s*([^}]+[^\s])\s?\s*}}")
LARGER_RUNNER_REGEX_LIST = re.compile(
    r"(windows|ubuntu)-(24.04|22.04|20.04|2019-2022)-(4|8|16|32|64)core-(16|32|64|128|256)gb"
)
MATRIX_KEY_EXTRACTION_REGEX = re.compile(r"{{\s*matrix\.([\w-]+)\s*}}")
STATIC_IF = re.compile(r"^(\$\{\{)?[A-Za-z0-9. ]+(\}\})?$")


@staticmethod
def starts_with_any(value: str, prefixes: list[str]) -> bool:
    """Returns True if 'value' starts with any of the provided prefixes."""
    return any(value.startswith(prefix) for prefix in prefixes)


@staticmethod
def process_matrix(job_def, runs_on):
    """Process case where runner is specified via matrix."""
    try:

        matrix_match = MATRIX_KEY_EXTRACTION_REGEX.search(runs_on)

        if matrix_match:
            matrix_key = matrix_match.group(1)
        else:
            return False
        # Check if strategy exists in the yaml file
        if "strategy" in job_def and "matrix" in job_def["strategy"]:
            matrix = job_def["strategy"]["matrix"]
            # Fail open in case we have step output or fromJSON
            if type(matrix) is str:
                return True

            # Use previously acquired key to retrieve list of OSes
            if matrix_key in matrix:
                os_list = matrix[matrix_key]
            elif "include" in matrix:
                inclusions = matrix["include"]
                os_list = []
                for inclusion in inclusions:
                    if matrix_key in inclusion:
                        os_list.append(inclusion[matrix_key])
            else:
                return False

            # We only need ONE to be self hosted, others can be
            # GitHub hosted
            for key in os_list:
                if type(key) is str:
                    if key not in ConfigurationManager().WORKFLOW_PARSING[
                        "GITHUB_HOSTED_LABELS"
                    ] and not LARGER_RUNNER_REGEX_LIST.match(key):
                        return True
                # list of labels
                elif type(key) is list:
                    return True
    except TypeError as e:
        print("Error processing matrix job")
        print(job_def)
        return False


@staticmethod
def is_within_last_day(timestamp_str, format="%Y-%m-%dT%H:%M:%SZ"):
    # Convert the timestamp string to a datetime object
    date = datetime.strptime(timestamp_str, format)

    # Get the current date and time
    now = datetime.now()
    # Calculate the date 1 days ago
    one_day_ago = now - timedelta(days=1)

    # Return True if the date is within the last day, False otherwise
    return one_day_ago <= date <= now


@staticmethod
def return_recent(time1, time2, format="%Y-%m-%dT%H:%M:%SZ"):
    """
    Takes two timestamp strings and returns the most recent one.

    Args:
        time1 (str): The first timestamp string.
        time2 (str): The second timestamp string.
        format (str): The format of the timestamp strings. Default is '%Y-%m-%dT%H:%M:%SZ'.

    Returns:
        str: The most recent timestamp string.
    """
    # Convert the timestamp strings to datetime objects
    date1 = datetime.strptime(time1, format)
    date2 = datetime.strptime(time2, format)

    # Return the most recent timestamp string
    return time1 if date1 > date2 else time2


@staticmethod
def process_runner(runs_on):
    """
    Processes the runner for the job.
    """
    if type(runs_on) is list:
        for label in runs_on:
            if label in ConfigurationManager().WORKFLOW_PARSING["GITHUB_HOSTED_LABELS"]:
                break
            if LARGER_RUNNER_REGEX_LIST.match(label):
                break
        else:
            return True
    elif type(runs_on) is str:
        if runs_on in ConfigurationManager().WORKFLOW_PARSING["GITHUB_HOSTED_LABELS"]:
            return False
        if LARGER_RUNNER_REGEX_LIST.match(runs_on):
            return False
        return True

    return False


@staticmethod
def parse_script(contents: str):
    """Processes run steps for additional context"""
    return_dict = {
        "is_checkout": False,
        "metadata": None,
        "is_sink": False,
        "hard_gate": False,
        "soft_gate": False,
    }

    if not contents or not isinstance(contents, str):
        logging.warning("Invalid contents for script parsing:" + str(contents))
        return return_dict

    if "git checkout" in contents or "pr checkout" in contents:
        match = pattern.search(contents)
        if match:
            ref = match.group(2)

            static_vals = ["base", "main", "master"]

            for prefix in ConfigurationManager().WORKFLOW_PARSING["PR_ISH_VALUES"]:
                if prefix in ref.lower() and not any(
                    substring in ref.lower() for substring in static_vals
                ):
                    return_dict["metadata"] = ref
                    return_dict["is_checkout"] = True

    if "isCrossRepository" in contents and "GITHUB_OUTPUT" in contents:
        return_dict["hard_gate"] = True
    if "github.rest.repos.checkCollaborator" in contents:
        return_dict["soft_gate"] = True
    if "getCollaboratorPermissionLevel" in contents:
        return_dict["soft_gate"] = True
    if "getMembershipForUserInOrg" in contents:
        return_dict["soft_gate"] = True

    if check_sinks(contents):
        return_dict["is_sink"] = True
    return return_dict


@staticmethod
def check_sinks(script):
    """Check if the script contain a sink."""
    sinks = ConfigurationManager().WORKFLOW_PARSING["SINKS"]
    sinks_start = ConfigurationManager().WORKFLOW_PARSING["SINKS_START"]

    for sink in sinks:
        if sink in script:
            return True

    lines = script.splitlines()
    for line in lines:
        for sink in sinks_start:
            if line.strip().startswith(sink):
                return True
    return False


@staticmethod
def check_sus(item):
    """
    Check if the given item starts with any of the predefined potentially problematic prefixes.

    This method is used to identify potentially unsafe or
    suspicious variables in a GitHub Actions workflow.
    It checks if the item starts with any of the prefixes
    defined in PREFIX_VALUES. These prefixes are typically
    used to reference variables in a GitHub Actions workflow,
    and if a user-controlled variable is referenced
    without proper sanitization, it could lead to a script
    injection vulnerability.

    Args:
        item (str): The item to check.

    Returns:
        bool: True if the item starts with any of the suspicious prefixes, False otherwise.
    """

    PREFIX_VALUES = ["needs.", "env.", "steps.", "jobs."]

    item_lower = item.lower()
    for prefix in PREFIX_VALUES:
        if item_lower.startswith(prefix):
            for safe_string in ConfigurationManager().WORKFLOW_PARSING[
                "SAFE_ISH_CONTEXTS"
            ]:
                if item_lower.endswith(safe_string):
                    break
            else:
                return True
    return False


@staticmethod
def check_risky_regexes(item):
    regexes = ConfigurationManager().WORKFLOW_PARSING["RISKY_CONTEXT_REGEXES"]

    for regex in regexes:
        if re.search(regex, item):
            return True


@staticmethod
def check_pr_ref(item):
    """
    Checks if the given item contains any of the predefined pull request
    related values.

    This method is used to identify if a given item (typically a string)
    contains any of the values defined in PR_ISH_VALUES. These values are
    typically used to reference pull request related data in a GitHub Actions workflow.

    Args:
        item (str): The item to check.

    Returns:
        bool: True if the item contains any of the pull request related values, False otherwise.
    """
    for prefix in ConfigurationManager().WORKFLOW_PARSING["PR_ISH_VALUES"]:
        if prefix in item.lower() and "base" not in item.lower():
            return True
    return False


@staticmethod
def filter_tokens(tokens, strict=False):
    """
    This method filters the tokens from the contents of a step
    in a GitHub workflow for potentially unsafe or suspicious context expressions.

    Parameters:
    contents (str): The contents of a step in a GitHub workflow.

    The method uses a regular expression to find all context expressions in the contents.
    A context expression in a GitHub workflow is a string that starts with '${{' and ends with '}}'.
    The expression inside the curly braces is split by '.' into three parts.
    The first part must be a sequence of alphanumeric characters,
    the second part must also be a sequence of alphanumeric characters,
    and the third part can be any sequence of characters.

    The method then checks each found context expression in two ways:
    1. Against a list of known unsafe context expressions.
        If the context expression is found in the list, it is added to `tokens_knownbad`.
    2. Using the `check_sus` method. If `check_sus` returns
        True for a context expression, it is added to `tokens_sus`.

    The method returns a list of tokens that are either known to be unsafe or are considered suspicious.

    Returns:
    list: A list of unsafe or suspicious context expressions found in the contents.
    """
    # First we get known unsafe
    tokens_knownbad = [
        item
        for item in tokens
        if item.lower() in ConfigurationManager().WORKFLOW_PARSING["UNSAFE_CONTEXTS"]
    ]
    # And then we add anything referenced
    if not strict:
        tokens_sus = [item for item in tokens if check_sus(item)]
        tokens_knownbad.extend(tokens_sus)

    return list(set(tokens_knownbad))


@staticmethod
def getToken(contents):
    """Get the context token from the step."""
    match = CONTEXT_REGEX.search(contents)
    if match:
        return match.group(1).replace(" ", "")
    return contents


@staticmethod
def checkUnsafe(variable):
    """Check if the variable is unsafe."""
    return variable in ConfigurationManager().WORKFLOW_PARSING["UNSAFE_CONTEXTS"]


@staticmethod
def prReviewUnsafe(variable):
    """Check if the variable is unsafe."""
    return (
        variable in ConfigurationManager().WORKFLOW_PARSING["PR_REVIEW_UNSAFE_CONTEXTS"]
    )


@staticmethod
def getTokens(contents):
    """Get the context tokens from the step."""
    if contents:
        finds = CONTEXT_REGEX.findall(contents)

        extension = None
        for find in finds:
            if " || " in find:
                extension = find.split(" || ")
                break

        if extension:
            finds.extend(extension)
        return finds
    else:
        return []


@staticmethod
def check_always_true(if_check):
    """Check if the if check is always true because it uses ${{ }} in combination with
    an expression or uses nested expressions.
    """
    # If it starts with an expression but ends without, then it
    # always will resolve to true.
    if if_check.strip().startswith("${{") and not if_check.endswith("}}"):
        return True


@staticmethod
def check_false_if(if_check):
    checks = ConfigurationManager().WORKFLOW_PARSING["SAFE_IF_CHECKS"]

    for check in checks:
        if check in if_check:
            return True


@staticmethod
def validate_if_check(if_check, variables={}):
    """Function used to validate each if check.
    The strategy here is to "fail open". If we cannot
    determine with certainty the check would fail in an injection
    or pwn request scenario then we return True, because we would rather
    not miss a potential vulnerability.
    """
    if not if_check:
        return True

    if check_false_if(if_check):
        return False

    if check_always_true(if_check):
        return True

    try:
        parser = ExpressionParser(if_check)
    except Exception:
        # Fail open
        return True
    ast_root = parser.get_node()

    try:
        evaluator = ExpressionEvaluator(variables)
        result = evaluator.evaluate(ast_root)
    except NotImplementedError:
        return True
    except Exception:
        return True

    return result


@staticmethod
def __is_within_last_day(timestamp_str, format="%Y-%m-%dT%H:%M:%SZ"):
    # Convert the timestamp string to a datetime object
    date = datetime.strptime(timestamp_str, format)

    # Get the current date and time
    now = datetime.now()
    # Calculate the date 1 days ago
    one_day_ago = now - timedelta(days=1)

    # Return True if the date is within the last day, False otherwise
    return one_day_ago <= date <= now


@staticmethod
def decompose_action_ref(action_path: str, repo_name: str):
    """ """
    action_parts = {
        "key": action_path,
        "path": action_path.split("@")[0] if "@" in action_path else action_path,
        "ref": action_path.split("@")[1] if "@" in action_path else "",
        "local": action_path.startswith("./"),
        "docker": False,
    }

    if "docker://" in action_path:
        # Gato-X doesn't support docker actions
        # and we ignore official GitHub actions for analysis.
        action_parts["docker"] = True
        return action_parts

    if not action_parts["local"]:
        path_parts = action_parts["path"].split("/")

        action_parts["repo"] = "/".join(path_parts[0:2])
        if len(path_parts) > 2:
            action_parts["path"] = "/".join(action_parts["path"].split("/")[2:])
        else:
            # Standard action path in base directory
            action_parts["path"] = ""
    else:
        action_parts["path"] = action_parts["path"][2:]
        action_parts["repo"] = repo_name

    return action_parts


@staticmethod
def parse_github_path(path):
    parts = path.split("@")
    ref = parts[1] if len(parts) > 1 else "main"
    repo_path = parts[0].split("/")
    repo_slug = "/".join(repo_path[0:2])
    file_path = "/".join(repo_path[2:]) if len(repo_path) > 1 else ""

    return repo_slug, file_path, ref



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.configuration.configuration_manager import ConfigurationManager
from gatox.caching.cache_manager import CacheManager
from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.issue_type import IssueType
from gatox.enumerate.results.complexity import Complexity
from gatox.enumerate.results.result_factory import ResultFactory
from gatox.workflow_graph.graph_builder import WorkflowGraphBuilder
from gatox.github.api import Api

from gatox.workflow_parser.utility import (
    CONTEXT_REGEX,
    is_within_last_day,
    return_recent,
)
from gatox.notifications.send_webhook import send_slack_webhook


class VisitorUtils:
    """Class to track contextual information during a single visit."""

    @staticmethod
    def _add_results(
        path,
        results: dict,
        issue_type,
        confidence: Confidence = Confidence.UNKNOWN,
        complexity: Complexity = Complexity.ZERO_CLICK,
    ):
        repo_name = path[0].repo_name()
        if repo_name not in results:
            results[repo_name] = []

        if issue_type == IssueType.ACTIONS_INJECTION:
            result = ResultFactory.create_injection_result(path, confidence, complexity)
        elif issue_type == IssueType.PWN_REQUEST:
            result = ResultFactory.create_pwn_result(path, confidence, complexity)
        elif issue_type == IssueType.DISPATCH_TOCTOU:
            result = ResultFactory.create_toctou_result(path, confidence, complexity)
        elif issue_type == IssueType.PR_REVIEW_INJECTON:
            result = ResultFactory.create_review_injection_result(
                path, confidence, complexity
            )
        else:
            raise ValueError(f"Unknown issue type: {issue_type}")

        results[repo_name].append(result)

    @staticmethod
    def initialize_action_node(graph, api, node):
        """
        Initialize an action node by removing the 'uninitialized' tag and setting it up.

        Args:
            graph (TaggedGraph):
                The workflow graph containing all nodes.
            api (Api):
                An instance of the API wrapper to interact with external services.
            node (Node):
                The node to be initialized.

        Returns:
            None

        Raises:
            None
        """
        tags = node.get_tags()
        if "uninitialized" in tags:
            WorkflowGraphBuilder()._initialize_action_node(node, api)
            graph.remove_tags_from_node(node, ["uninitialized"])

    @staticmethod
    def check_mutable_ref(ref, start_tags=set()):
        """
        Check if a reference is mutable based on allowed GitHub SHA patterns.

        Args:
            ref (str):
                The reference string to check.
            start_tags (set, optional):
                A set of starting tags for additional context. Defaults to an empty set.

        Returns:
            bool:
                False if the reference is immutable, True otherwise.
        """
        if "github.event.pull_request.head.sha" in ref:
            return False
        elif "github.event.workflow_run.head.sha" in ref:
            return False
        elif "github.sha" in ref:
            return False
        # If the trigger is pull_request_target and we have a sha in the reference, then this is very likely
        # to be from the original trigger in some form and not a mutable reference, so if it is gated we can suppress.
        elif "sha" in ref and "pull_request_target" in start_tags:
            return False
        # This points to the base branch, so it is not going to be exploitable.
        elif "github.ref" in ref and "||" not in ref:
            return False

        return True

    @staticmethod
    def process_context_var(value):
        """
        Process a context variable by extracting relevant parts.

        Args:
            value (str):
                The context variable string to process.

        Returns:
            str:
                The processed variable.
        """
        processed_var = value
        if "${{" in value:
            processed_var = CONTEXT_REGEX.findall(value)
            if processed_var:
                processed_var = processed_var[0]
                if "inputs." in processed_var:
                    processed_var = processed_var.replace("inputs.", "")
            else:
                processed_var = value
        else:
            processed_var = value
        return processed_var

    @staticmethod
    def append_path(head, tail):
        """
        Append the tail to the head if the tail starts with the last element of the head.

        Args:
            head (list):
                The initial path list.
            tail (list):
                The path to append.

        Returns:
            list:
                The combined path if conditions are met; otherwise, the original head.
        """
        if head and tail and head[-1] == tail[0]:
            head.extend(tail[1:])
        return head

    @staticmethod
    def add_repo_results(data: dict, api: Api):
        """Add results to the repository data."""
        seen = set()
        for _, flows in data.items():
            for flow in flows:
                seen_before = flow.get_first_and_last_hash()
                if not seen_before in seen:
                    seen.add(seen_before)
                else:
                    continue

                repo = CacheManager().get_repository(flow.repo_name())
                repo.set_results(flow)

                if (
                    ConfigurationManager().NOTIFICATIONS["SLACK_WEBHOOKS"]
                    and repo
                    and is_within_last_day(repo.repo_data["pushed_at"])
                ):
                    value = flow.to_machine()
                    commit_date, author, sha = api.get_file_last_updated(
                        flow.repo_name(),
                        ".github/workflows/" + value.get("initial_workflow"),
                    )

                    merge_date = api.get_commit_merge_date(flow.repo_name(), sha)
                    if merge_date:
                        # If there is a PR merged, get the most recent.
                        commit_date = return_recent(commit_date, merge_date)

                    if is_within_last_day(commit_date) and "[bot]" not in author:
                        send_slack_webhook(value)



import pytest
import os
import pathlib
import requests  # If using the requests library

from unittest import mock
from unittest.mock import patch

from gatox.cli import cli
from gatox.util.arg_utils import read_file_and_validate_lines, is_valid_directory


@pytest.fixture(autouse=True)
def block_network_calls(monkeypatch):
    """
    Fixture to block real network calls during tests,
    raising an error if any attempt to send a request is made.
    """

    def mock_request(*args, **kwargs):
        raise RuntimeError("Blocked a real network call during tests.")

    monkeypatch.setattr(requests.sessions.Session, "request", mock_request)


@pytest.fixture(autouse=True)
def mock_settings_env_vars():
    """
    Fixture that mocks the GH_TOKEN environment variable
    for all tests in this module.
    """
    with mock.patch.dict(
        os.environ, {"GH_TOKEN": "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"}
    ):
        yield


@patch("builtins.input", return_value="")
def test_cli_no_gh_token(mock_input, capfd):
    """Test case where no GH Token is provided"""
    del os.environ["GH_TOKEN"]

    with pytest.raises(SystemExit):
        cli.cli(["enumerate", "-t", "test"])

    mock_input.assert_called_with(
        "No 'GH_TOKEN' environment variable set! Please enter a GitHub" " PAT.\n"
    )


def test_cli_fine_grained_pat(capfd):
    """Test case where an unsupported PAT is provided."""
    os.environ["GH_TOKEN"] = "github_pat_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

    with pytest.raises(SystemExit):
        cli.cli(["enumerate", "-t", "test"])
    out, err = capfd.readouterr()
    assert "not supported" in err


def test_cli_s2s_token(capfd):
    """Test case where a service-to-service token is provided."""
    os.environ["GH_TOKEN"] = "ghs_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

    with pytest.raises(SystemExit):
        cli.cli(["enumerate", "-t", "test"])
    out, err = capfd.readouterr()
    assert "not support App tokens without machine flag" in err


def test_cli_s2s_token_no_machine(capfd):
    """Test case where a service-to-service token is provided."""
    os.environ["GH_TOKEN"] = "ghs_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

    with pytest.raises(SystemExit):
        cli.cli(["enumerate", "-r", "testOrg/testRepo"])
    out, err = capfd.readouterr()
    assert "not support App tokens without machine flag" in err


@patch("gatox.enumerate.enumerate.Api.call_get")
@patch("gatox.enumerate.enumerate.Api.call_post")
def test_cli_s2s_token_machine(mock_post, mock_get, capfd):
    """Test case where a service-to-service token is provided."""
    import os
    from gatox.cli import cli  # [gatox/cli/cli.py](gatox/cli/cli.py)

    os.environ["GH_TOKEN"] = "ghs_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

    # Mock out the enumerator’s HTTP calls here as needed
    mock_get.return_value.status_code = 200
    mock_get.return_value.json.return_value = {"total_count": 0}
    mock_post.return_value.status_code = 200

    cli.cli(["enumerate", "-r", "testOrg/testRepo", "--machine"])
    out, _ = capfd.readouterr()
    assert "Allowing the use of a GitHub App token for single repo enumeration" in out


def test_cli_u2s_token(capfd):
    """Test case where a service-to-service token is provided."""
    os.environ["GH_TOKEN"] = "ghu_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

    with pytest.raises(SystemExit):
        cli.cli(["enumerate", "-t", "test"])
    out, err = capfd.readouterr()
    assert "Provided GitHub PAT is malformed or unsupported" in err


@mock.patch("gatox.cli.cli.Enumerator")
def test_cli_oauth_token(mock_enumerate, capfd):
    """Test case where a GitHub oauth token is provided."""
    os.environ["GH_TOKEN"] = "gho_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"

    mock_instance = mock_enumerate.return_value
    mock_api = mock.MagicMock()
    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }
    mock_api.get_user_type.return_value = "Organization"
    mock_instance.api = mock_api

    cli.cli(["enumerate", "-t", "test"])
    out, err = capfd.readouterr()

    mock_enumerate.return_value.enumerate_organization.assert_called_once()


@mock.patch("gatox.cli.cli.Enumerator")
def test_cli_old_token(mock_enumerate, capfd):
    """Test case where an old, but still potentially valid GitHub token is provided."""
    os.environ["GH_TOKEN"] = "43255147468edf32a206441ad296ce648f44ee32"

    mock_instance = mock_enumerate.return_value
    mock_api = mock.MagicMock()
    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }
    mock_api.get_user_type.return_value = "Organization"
    mock_instance.api = mock_api

    cli.cli(["enumerate", "-t", "test"])
    out, err = capfd.readouterr()

    mock_instance.enumerate_organization.assert_called_once()


def test_cli_invalid_pat(capfd):
    """Test case where a clearly invalid PAT is provided."""
    os.environ["GH_TOKEN"] = "invalid"

    with pytest.raises(SystemExit):
        cli.cli(["enumerate", "-t", "test"])
    out, err = capfd.readouterr()
    assert "malformed" in err


def test_cli_double_proxy(capfd):
    """Test case where conflicing proxies are provided."""
    with pytest.raises(SystemExit):
        cli.cli(["-sp", "socks", "-p", "http", "enumerate", "-t", "test"])

    out, err = capfd.readouterr()
    assert "proxy at the same time" in err


def test_attack_bad_args1(capfd):
    """Test attack command without the attack method."""

    with pytest.raises(SystemExit):
        cli.cli(["attack", "-t", "test"])

    out, err = capfd.readouterr()
    assert "must select one" in err


def test_attack_bad_args2(capfd):
    """Test attack command with conflicting params."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    with pytest.raises(SystemExit):
        cli.cli(
            [
                "attack",
                "-t",
                "test",
                "-pr",
                "-f",
                os.path.join(curr_path, "files/main.yaml"),
                "-n",
                "invalid",
            ]
        )

    out, err = capfd.readouterr()
    assert "cannot be used with a custom" in err


def test_attack_invalid_path(capfd):
    """Test attack command with an invalid path."""

    with pytest.raises(SystemExit):
        cli.cli(["attack", "-t", "test", "-pr", "-f", "path"])

    out, err = capfd.readouterr()
    assert "argument --custom-file/-f: The file: path does not exist!" in err


def test_repos_file_good():
    """Test that the good file is validated without errors."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    res = read_file_and_validate_lines(
        os.path.join(curr_path, "files/test_repos_good.txt"),
        r"[A-Za-z0-9-_.]+\/[A-Za-z0-9-_.]+",
    )

    assert "someorg/somerepository" in res
    assert "some_org/some-repo" in res


def test_repos_file_bad(capfd):
    """Test that the good file is validated without errors."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    with pytest.raises(SystemExit):
        cli.cli(
            ["enumerate", "-R", os.path.join(curr_path, "files/test_repos_bad.txt")]
        )

    out, err = capfd.readouterr()

    assert "invalid repository name!" in err


def test_valid_dir():
    """Test that the directory validation function works."""
    curr_path = pathlib.Path(__file__).parent.resolve()
    mock_parser = mock.MagicMock()

    res = is_valid_directory(mock_parser, os.path.join(curr_path, "files/"))

    assert res == os.path.join(curr_path, "files/")


def test_invalid_dir(capfd):
    """Test that the directory validation function works."""
    curr_path = pathlib.Path(__file__).parent.resolve()
    mock_parser = mock.MagicMock()

    res = is_valid_directory(mock_parser, os.path.join(curr_path, "invaliddir/"))

    assert res is None

    mock_parser.error.assert_called_with(
        "The directory {} does not exist!".format(
            os.path.join(curr_path, "invaliddir/")
        )
    )


@mock.patch("gatox.attack.runner.webshell.WebShell.runner_on_runner")
def test_attack_pr(mock_attack):
    """Test attack command using the pr method."""
    cli.cli(
        ["attack", "-t", "test", "-pr", "--target-os", "linux", "--target-arch", "x64"]
    )
    mock_attack.assert_called_once()


@mock.patch("gatox.attack.runner.webshell.WebShell.runner_on_runner")
def test_attack_pr_bados(mock_attack, capfd):
    """Test attack command using the pr method."""
    with pytest.raises(SystemExit):
        cli.cli(
            [
                "attack",
                "-t",
                "test",
                "-pr",
                "--target-os",
                "solaris",
                "--target-arch",
                "x64",
            ]
        )

    out, err = capfd.readouterr()
    assert "invalid choice: 'solaris'" in err


@mock.patch("gatox.attack.attack.Attacker.push_workflow_attack")
def test_attack_workflow(mock_attack):
    """Test attack command using the workflow method."""

    cli.cli(["attack", "-t", "test", "-w"])
    mock_attack.assert_called_once()


@mock.patch("os.path.isdir")
def test_enum_bad_args1(mock_dircheck, capfd):
    """Test enum command with invalid output location."""
    mock_dircheck.return_value = False

    with pytest.raises(SystemExit):
        cli.cli(["enum", "-o", "invalid"])

    out, err = capfd.readouterr()
    assert "--output-yaml/-o: The directory: invalid does not exist!" in err


def test_enum_bad_args2(capfd):
    """Test enum command without a type selection."""
    with pytest.raises(SystemExit):
        cli.cli(["enum"])

    out, err = capfd.readouterr()
    assert "type was specified" in err


def test_enum_bad_args3(capfd):
    """Test enum command with multiple type selections."""
    with pytest.raises(SystemExit):
        cli.cli(["enum", "-t", "test", "-r", "testorg/test2"])

    out, err = capfd.readouterr()
    assert "select one enumeration" in err


@mock.patch("gatox.enumerate.enumerate.Enumerator.self_enumeration")
def test_enum_self(mock_enumerate):
    """Test enum command using the self enumerattion."""

    mock_enumerate.return_value = [["org1"], ["org2"]]

    cli.cli(["enum", "-s"])
    mock_enumerate.assert_called_once()


@mock.patch("gatox.models.execution.Execution.add_repositories")
@mock.patch("gatox.models.execution.Execution.add_organizations")
@mock.patch("gatox.enumerate.enumerate.Enumerator.self_enumeration")
def test_enum_self_json_empty(mock_enumerate, mock_executor_org, mock_executor_repo):
    """Test enum command using the self enumerattion."""

    mock_enumerate.return_value = ([], ["repo1", "repo2"])

    cli.cli(["enum", "-s", "-oJ", "test.json"])
    mock_enumerate.assert_called_once()

    mock_executor_org.assert_called_with([])
    mock_executor_repo.assert_called_with(["repo1", "repo2"])


@mock.patch("gatox.cli.cli.Enumerator")
def test_enum_org(mock_enumerate):
    """Test enum command using the organization enumerattion."""

    mock_instance = mock_enumerate.return_value
    mock_api = mock.MagicMock()

    print(mock_instance)

    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }
    mock_api.get_user_type.return_value = "Organization"
    mock_instance.api = mock_api

    cli.cli(["enum", "-t", "test"])

    mock_instance.enumerate_organization.assert_called_once()


@mock.patch("gatox.cli.cli.Enumerator")
def test_enum_user(mock_enumerate):
    """Test enum command using the organization enumeration."""

    mock_instance = mock_enumerate.return_value
    mock_api = mock.MagicMock()

    mock_api.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }
    mock_api.get_user_type.return_value = "User"
    mock_instance.api = mock_api

    cli.cli(["enum", "-t", "testUser"])

    mock_instance.enumerate_user.assert_called_once()


@mock.patch("gatox.enumerate.enumerate.Enumerator.enumerate_repos")
@mock.patch("gatox.util.read_file_and_validate_lines")
def test_enum_repos(mock_read, mock_enumerate):
    """Test enum command using the repo list."""
    curr_path = pathlib.Path(__file__).parent.resolve()
    mock_read.return_value = "repos"

    cli.cli(["enum", "-R", os.path.join(curr_path, "files/test_repos_good.txt")])
    mock_read.assert_called_once()
    mock_enumerate.assert_called_once()


@mock.patch("gatox.enumerate.enumerate.Enumerator.enumerate_repos")
def test_enum_repo(mock_enumerate):
    """Test enum command using the organization enumerattion."""
    cli.cli(["enum", "-r", "testorg/testrepo"])
    mock_enumerate.assert_called_once()


@mock.patch("gatox.search.search.Searcher.use_search_api")
def test_search(mock_search):
    """Test search command"""

    cli.cli(["search", "-t", "test"])
    mock_search.assert_called_once()


def test_long_repo_name(capfd):
    """Test enum command using name that is too long."""

    repo_name = "Org/" + "A" * 80

    with pytest.raises(SystemExit):
        cli.cli(["enum", "-r", repo_name])

    out, err = capfd.readouterr()

    assert "The maximum length is 79 characters!" in err


def test_invalid_repo_name(capfd):
    """Test enum command using invalid full repo name."""
    with pytest.raises(SystemExit):
        cli.cli(["enum", "-r", "RepoWithoutOrg"])

    out, err = capfd.readouterr()

    assert (
        "argument --repository/-r: The argument" " is not in the valid format!" in err
    )


@mock.patch("gatox.util.arg_utils.os.access")
def test_unreadable_file(mock_access, capfd):
    """Test enum command unreadable file."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    mock_access.return_value = False

    with pytest.raises(SystemExit):
        cli.cli(["enum", "-R", os.path.join(curr_path, "files/bad_dir/bad_file")])

    out, err = capfd.readouterr()

    assert " is not readable" in err


@mock.patch("gatox.util.arg_utils.os.access")
def test_unwritable_dir(mock_access, capfd):
    """Test enum command unwriable dir."""
    curr_path = pathlib.Path(__file__).parent.resolve()

    mock_access.return_value = False

    with pytest.raises(SystemExit):
        cli.cli(
            [
                "enum",
                "-r",
                "testOrg/testRepo",
                "-o",
                os.path.join(curr_path, "files/bad_dir"),
            ]
        )

    out, err = capfd.readouterr()

    assert " is not writeable" in err



import time
import logging

from gatox.github.api import Api
from gatox.cli.output import Output

from urllib.parse import urlparse


logger = logging.getLogger(__name__)


class Search:
    """Search utility for GH api in order to find public repos that may have
    security issues.
    """

    def __init__(self, api_accessor: Api):
        """Initialize class to call GH search methods. Due to the late limiting
        associated with these API calls.


        Args:
            api_accesor (Api): API accesor to use when making GitHub
            API requests.
        """
        self.api_accessor = api_accessor

    def search_enumeration(self, organization: str = None, custom_query: str = None):
        """Search for self-hosted in yml files within a given organization.

        Args:
            organization (str): Name of the github organization.
            custom_query (str, optional): Optional query to override default.

        Returns:
            set: Set containing repositories that are of interest.
        """

        query = {"sort": "indexed", "per_page": "100", "page": 1}

        if custom_query:
            query["q"] = custom_query
        else:
            query["q"] = (
                f"self-hosted org:{organization} language:yaml "
                "path:.github/workflows"
            )

        next_page = (
            f"/search/code?q={query['q']}&sort={query['sort']}"
            f"&per_page={query['per_page']}&page={query['page']}"
        )

        Output.info("Searching", end="", flush=True)
        candidates = set()
        while next_page:
            result = self.api_accessor.call_get(next_page)
            print(".", end="", flush=True)
            code = result.status_code
            data = result.json()
            headers = result.headers

            if code == 403:
                retry_after = headers.get("retry-after")
                reset = headers.get("x-ratelimit-reset")
                sleep = 60
                if retry_after:
                    sleep = int(retry_after) + 5
                elif reset:
                    sleep = int(reset) - int(time.time()) + 5

                print()
                Output.warn(
                    f"Secondary API Rate Limit Hit. Sleeping for {sleep} seconds!"
                )
                time.sleep(sleep)

                Output.info("Searching", end="", flush=True)
                continue
            elif code != 200:
                print()
                Output.error(f"Search failed with response code {code}!")

                context = result.json()
                if "errors" in context and len(context["errors"]) > 0:
                    Output.warn(
                        "\tError message from GitHub:\n"
                        f"\t{context['errors'][0]['message']}"
                    )

                return candidates

            if "incomplete_results" in data and data["incomplete_results"]:
                print()
                Output.warn("Search results incomplete due to GitHub timeout!")

            for entry in data["items"]:
                candidates.add(entry["repository"]["full_name"])

            next_page = result.links.get("next", {}).get("url")
            if next_page:
                link = urlparse(next_page)
                next_page = f"{link.path}?{link.query}"
                time.sleep(5)

        print()
        return candidates
